#include "..\..\common\ImageFilter.h"
#include "SettingsForm.h"
#include <set>

using namespace FilterSpace;
using namespace System::Windows::Forms;
using namespace std;

#define DOUBLESHORTINT_TO_INT(a,b)     ((a) << 16 | ((b) & 0xffff))
#define FIRST_OF_DOUBLESHORTINT(a)     ((a) >> 16)
#define SECOND_OF_DOUBLESHORTINT(a)    ((a) & 0xffff)

int recognition_directions[2][8] = { { 1, 1, 0, -1, -1, -1, 0, 1 }, { 0, 1, 1, 1, 0, -1, -1, -1 } };




bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// WRITE YOUR FILTER CODE HERE -->
	int   blob_index_remove = f->blob_index_remove;
	int   blob_index_new    = f->blob_index_new;
	int   blob_size_limit   = f->blob_index_size_limit;
	int   width             = f->width;
	int   height            = f->height;
	int   stride            = f->stridein;
	int   mstride           = (width + 7) >> 3;
	byte *din               = f->datain;
	byte *dout              = f->dataout;
	byte *identified        = new byte[mstride * height];

	memset(identified, 0, mstride * height);

	if (din != dout) memcpy(dout, din, stride * height);

	unsigned int coord;
	int	nx, ny, fx, fy, cnt;
	set<unsigned int> blob_pixels_to_extract;
	set<unsigned int> blob_pixels;

	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
		{
			// AZ AKTULIS BLOB FELDOLGOZSA
			cnt = 0;
			if ((identified[y * mstride + (x >> 3)] & (1 << (x & 7))) == 0 && din[y * stride + x] == blob_index_remove)
			{
				// BLOB SZNNEK S EGY PONTJNAK KIVLASZTSA
				coord = DOUBLESHORTINT_TO_INT(y,x);

				blob_pixels_to_extract.insert(coord);
				blob_pixels.insert(coord);
				identified[y * mstride + (x >> 3)] |= (1 << (x & 7));

				// A BLOB KITERJESZTSE
				while (blob_pixels_to_extract.size())
				{
					coord = *(blob_pixels_to_extract.begin());
					blob_pixels_to_extract.erase(coord);
					fy = FIRST_OF_DOUBLESHORTINT(coord);
					fx = SECOND_OF_DOUBLESHORTINT(coord);

					// BLOB KITERJESZTSE
					for (int i = 0; i < 8; i+=2)
					{
						nx = fx + recognition_directions[0][i];
						ny = fy + recognition_directions[1][i];
						if (nx < 0 || ny < 0 || nx > width - 1 || ny > height - 1) continue;

						if (din[ny * stride + nx] == blob_index_remove && (identified[ny * mstride + (nx >> 3)] & (1 << (nx & 7))) == 0) 
						{
							coord = DOUBLESHORTINT_TO_INT(ny,nx);
							blob_pixels_to_extract.insert(coord);

							if (cnt < blob_size_limit)
							{
								cnt++;
								blob_pixels.insert(coord);
							}
							else
							{
								cnt = 0x0fffffff;
							}
							identified[ny * mstride + (nx >> 3)] |= (1 << (nx & 7));
						}
					}
				}

				if (cnt != 0x0fffffff)
				{
					for (set<unsigned int>::iterator it = blob_pixels.begin(); it != blob_pixels.end(); it++)
					{
						coord = *it;
						fy = FIRST_OF_DOUBLESHORTINT(coord);
						fx = SECOND_OF_DOUBLESHORTINT(coord);
						dout[fy * stride + fx] = blob_index_new;
					}
				}
				blob_pixels.clear();
			}
			else
			{
				identified[y * mstride + (x >> 3)] |= (1 << (x & 7));
			}
		}


		// FOR PERFORMANCE MEASUREMENT
		ImageFilterBase::UpdateFilterStatus(f, width, updater);
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}
	delete [] identified;

	return true;
}




IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int indata = 0, outdata = 0, stride = 0, height = 0, run_local = 1;

	if (!fs) return IntPtr(0);

	bool cancelled = false;
	xml_node^ xn = gcnew xml_node( fs );

	filtersettings* f = XmlToMyStruct( xn, true );

	if (f && !silent)
	{
		FilterTemplate::SettingsForm^ form = gcnew FilterTemplate::SettingsForm(f);
		form->ShowDialog();
		cancelled = form->FilterCancelled;
		delete form;
	}

	if (cancelled)
	{
		delete f;
		delete xn;
		return IntPtr(0);
	}

	// GET "Run_Local" PARAMETER
	xn->XmlGetChildValueByName("Run_Local", run_local);

	// CLONE "InputImage0" TO "OutputImage0"
	xml_node^ xt = nullptr;

	if (xn->XmlGetChildByName(xt, "InputImage0" )) xt->XmlGetChildValueByName("Data", indata);
	if (xn->XmlGetChildByName(xt, "OutputImage0")) xt->XmlGetChildValueByName("Data", outdata);

	xn->XmlRemoveChildByName("OutputImage0");
	xn->XmlGetChildByName(xt, "InputImage0");

	Byte* fb = xt->XML_Tree_Marshal();
	xt = gcnew xml_node(fb);
	delete [] fb;
	xt->name = "OutputImage0";

	// SET "Run_Local" settings
	if (!run_local)
	{
		if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
		{
			xt->XmlGetChildValueByName("Stride", stride);
			xt->XmlGetChildValueByName("Height", height);
			if (stride * height == 0) return IntPtr(0);
			xt->XmlSetChildValueByName("Data", ((int)(void*)(new Byte[stride * height])));
		}
		else
		{
			xt->XmlSetChildValueByName("Data", outdata);
		}
	}
	xn->children.Add( xt );

	xn->XmlSetChildValueByName("BlobIndexRemove",     f->blob_index_remove);
	xn->XmlSetChildValueByName("BlobIndexNew",        f->blob_index_new);
	xn->XmlSetChildValueByName("BlobSizeLimit",       f->blob_index_size_limit);
	xn->XmlSetChildValueByName("nCore!ExecutionSize", 1); // PIXELWISE == 1 ALWAYS FOR THIS FILTER

	delete f;
	Byte* res = xn->XML_Tree_Marshal();
	delete xn;
	return (IntPtr)(void*)res;
}


filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	xml_node^ xntmp, ^xt2;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;

	int packed, interleaved;
	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain)) return 0;

	if (xntmp->XmlGetChildByName(xt2, "ChannelsBitAllocation"))
	{
		String^ colMap;
		if (xt2->XmlGetChildValueByName("ColorMap", colMap))
		{
			unsigned char* data;
			int length;
			base64(StringTostring(colMap), data, length);
			memcpy(filset->palette256, data, MIN(1024,length));
		}
	} 

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->strideout, packed, interleaved, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	xn->XmlSetChildValueByName("BlobIndexRemove", 0,   true);
	xn->XmlGetChildValueByName("BlobIndexRemove", filset->blob_index_remove);
	xn->XmlSetChildValueByName("BlobIndexNew",    0, true);
	xn->XmlGetChildValueByName("BlobIndexNew",    filset->blob_index_new);
	xn->XmlSetChildValueByName("BlobSizeLimit",   0, true);
	xn->XmlGetChildValueByName("BlobSizeLimit",   filset->blob_index_size_limit);

	//MessageBox::Show("remove: " + filset->blob_index_remove.ToString() + "\nnew: " + filset->blob_index_new.ToString() + "\nlimit: " + filset->blob_index_size_limit.ToString());

	filset->processing_size    = ( xn->XmlGetChildValueByName("Processing_Size", m) ) ? (int*)(void*)m : 0;
	filset->processing_ptr     = ( xn->XmlGetChildValueByName("Processing_Ptr", m) ) ? (int*)(void*)m : 0;
	filset->nCoreEnabled       = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal    = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) ) ? m : 1;

	return filset;
}



bool ImageFilter::CanLocalize(){ return true; }


IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName,  width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}


// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		*(filset->processing_ptr)	= 0;
		*(filset->processing_size)	= filset->width * filset->height;
		b = (::FilterCore(filset, 0, 0, filset->width, filset->height, nullptr));
		delete filset;
		*(filset->processing_ptr)  = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}
